@@ -8,6 +8,26 @@ pub type Result<T> = std::result::Result<T, Error>;
 type Cause = Box<dyn StdError + Send + Sync>;
 
 /// Represents errors that can occur handling HTTP streams.
+///
+/// # Formatting
+///
+/// The `Display` implementation of this type will only print the details of
+/// this level of error, even though it may have been caused by another error
+/// and contain that error in its source. To print all the relevant
+/// information, including the source chain, using something like
+/// `std::error::Report`, or equivalent 3rd party types.
+///
+/// The contents of the formatted error message of this specific `Error` type
+/// is unspecified. **You must not depend on it.** The wording and details may
+/// change in any version, with the goal of improving error messages.
+///
+/// # Source
+///
+/// A `hyper::Error` may be caused by another error. To aid in debugging,
+/// those are exposed in `Error::source()` as erased types. While it is
+/// possible to check the exact type of the sources, they **can not be depended
+/// on**. They may come from private internal dependencies, and are subject to
+/// change at any moment.
 pub struct Error {
     inner: Box<ErrorImpl>,
 }
@@ -22,85 +42,99 @@ pub(super) enum Kind {
     Parse(Parse),
     User(User),
     /// A message reached EOF, but is not complete.
-    #[allow(unused)]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     IncompleteMessage,
     /// A connection received a message (or bytes) when not waiting for one.
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     UnexpectedMessage,
     /// A pending item was dropped before ever being processed.
     Canceled,
     /// Indicates a channel (client or body sender) is closed.
+    #[cfg(any(
+        all(feature = "http1", any(feature = "client", feature = "server")),
+        all(feature = "http2", feature = "client")
+    ))]
     ChannelClosed,
     /// An `io::Error` that occurred while trying to read or write to a network stream.
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    Io,
-    /// Error occurred while connecting.
-    #[allow(unused)]
-    Connect,
-    /// Error creating a TcpListener.
     #[cfg(all(
-        any(feature = "http1", feature = "http2"),
-        feature = "tcp",
-        feature = "server"
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
     ))]
-    Listen,
-    /// Error accepting on an Incoming stream.
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "server")]
-    Accept,
+    Io,
+    /// User took too long to send headers
+    #[cfg(all(feature = "http1", feature = "server"))]
+    HeaderTimeout,
     /// Error while reading a body from connection.
-    #[cfg(any(feature = "http1", feature = "http2", feature = "stream"))]
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     Body,
     /// Error while writing a body to connection.
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     BodyWrite,
-    /// The body write was aborted.
-    BodyWriteAborted,
     /// Error calling AsyncWrite::shutdown()
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     Shutdown,
 
     /// A general error from h2.
-    #[cfg(feature = "http2")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http2"))]
     Http2,
 }
 
 #[derive(Debug)]
 pub(super) enum Parse {
     Method,
-    Version,
     #[cfg(feature = "http1")]
+    Version,
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     VersionH2,
     Uri,
+    #[cfg(all(feature = "http1", feature = "server"))]
+    UriTooLong,
+    #[cfg(feature = "http1")]
     Header(Header),
+    #[cfg(any(feature = "http1", feature = "http2"))]
     TooLarge,
     Status,
-    #[cfg_attr(debug_assertions, allow(unused))]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     Internal,
 }
 
 #[derive(Debug)]
+#[cfg(feature = "http1")]
 pub(super) enum Header {
     Token,
-    #[cfg(feature = "http1")]
+    #[cfg(any(feature = "client", feature = "server"))]
     ContentLengthInvalid,
-    #[cfg(feature = "http1")]
+    #[cfg(feature = "server")]
     TransferEncodingInvalid,
-    #[cfg(feature = "http1")]
+    #[cfg(any(feature = "client", feature = "server"))]
     TransferEncodingUnexpected,
 }
 
 #[derive(Debug)]
 pub(super) enum User {
-    /// Error calling user's HttpBody::poll_data().
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    /// Error calling user's Body::poll_data().
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     Body,
-    /// Error calling user's MakeService.
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "server")]
-    MakeService,
+    /// The user aborted writing of the outgoing body.
+    #[cfg(any(
+        all(feature = "http1", any(feature = "client", feature = "server")),
+        feature = "ffi"
+    ))]
+    BodyWriteAborted,
     /// Error from future of user's Service.
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    #[cfg(any(
+        all(any(feature = "client", feature = "server"), feature = "http1"),
+        all(feature = "server", feature = "http2")
+    ))]
     Service,
     /// User tried to send a certain header in an unexpected context.
     ///
@@ -108,30 +142,22 @@ pub(super) enum User {
     #[cfg(any(feature = "http1", feature = "http2"))]
     #[cfg(feature = "server")]
     UnexpectedHeader,
-    /// User tried to create a Request with bad version.
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    UnsupportedVersion,
-    /// User tried to create a CONNECT Request with the Client.
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    UnsupportedRequestMethod,
     /// User tried to respond with a 1xx (not 101) response code.
     #[cfg(feature = "http1")]
     #[cfg(feature = "server")]
     UnsupportedStatusCode,
-    /// User tried to send a Request with Client with non-absolute URI.
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    AbsoluteUriRequired,
 
     /// User tried polling for an upgrade that doesn't exist.
     NoUpgrade,
 
     /// User polled for an upgrade, but low-level API is not using upgrades.
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     ManualUpgrade,
 
+    /// The dispatch task is gone.
+    #[cfg(all(feature = "client", any(feature = "http1", feature = "http2")))]
+    DispatchGone,
+
     /// User aborted in an FFI callback.
     #[cfg(feature = "ffi")]
     AbortedByCallback,
@@ -148,8 +174,12 @@ impl Error {
     }
 
     /// Returns true if this was an HTTP parse error caused by a message that was too large.
+    #[cfg(all(feature = "http1", feature = "server"))]
     pub fn is_parse_too_large(&self) -> bool {
-        matches!(self.inner.kind, Kind::Parse(Parse::TooLarge))
+        matches!(
+            self.inner.kind,
+            Kind::Parse(Parse::TooLarge) | Kind::Parse(Parse::UriTooLong)
+        )
     }
 
     /// Returns true if this was an HTTP parse error caused by an invalid response status code or
@@ -170,22 +200,41 @@ impl Error {
 
     /// Returns true if a sender's channel is closed.
     pub fn is_closed(&self) -> bool {
-        matches!(self.inner.kind, Kind::ChannelClosed)
-    }
+        #[cfg(not(any(
+            all(feature = "http1", any(feature = "client", feature = "server")),
+            all(feature = "http2", feature = "client")
+        )))]
+        return false;
 
-    /// Returns true if this was an error from `Connect`.
-    pub fn is_connect(&self) -> bool {
-        matches!(self.inner.kind, Kind::Connect)
+        #[cfg(any(
+            all(feature = "http1", any(feature = "client", feature = "server")),
+            all(feature = "http2", feature = "client")
+        ))]
+        matches!(self.inner.kind, Kind::ChannelClosed)
     }
 
     /// Returns true if the connection closed before a message could complete.
     pub fn is_incomplete_message(&self) -> bool {
+        #[cfg(not(all(any(feature = "client", feature = "server"), feature = "http1")))]
+        return false;
+
+        #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
         matches!(self.inner.kind, Kind::IncompleteMessage)
     }
 
     /// Returns true if the body write was aborted.
     pub fn is_body_write_aborted(&self) -> bool {
-        matches!(self.inner.kind, Kind::BodyWriteAborted)
+        #[cfg(not(any(
+            all(feature = "http1", any(feature = "client", feature = "server")),
+            feature = "ffi"
+        )))]
+        return false;
+
+        #[cfg(any(
+            all(feature = "http1", any(feature = "client", feature = "server")),
+            feature = "ffi"
+        ))]
+        matches!(self.inner.kind, Kind::User(User::BodyWriteAborted))
     }
 
     /// Returns true if the error was caused by a timeout.
@@ -193,11 +242,6 @@ impl Error {
         self.find_source::<TimedOut>().is_some()
     }
 
-    /// Consumes the error, returning its cause.
-    pub fn into_cause(self) -> Option<Box<dyn StdError + Send + Sync>> {
-        self.inner.cause
-    }
-
     pub(super) fn new(kind: Kind) -> Error {
         Error {
             inner: Box::new(ErrorImpl { kind, cause: None }),
@@ -214,10 +258,10 @@ impl Error {
         &self.inner.kind
     }
 
-    fn find_source<E: StdError + 'static>(&self) -> Option<&E> {
+    pub(crate) fn find_source<E: StdError + 'static>(&self) -> Option<&E> {
         let mut cause = self.source();
         while let Some(err) = cause {
-            if let Some(ref typed) = err.downcast_ref() {
+            if let Some(typed) = err.downcast_ref() {
                 return Some(typed);
             }
             cause = err.source();
@@ -227,7 +271,7 @@ impl Error {
         None
     }
 
-    #[cfg(feature = "http2")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http2"))]
     pub(super) fn h2_reason(&self) -> h2::Reason {
         // Find an h2::Reason somewhere in the cause stack, if it exists,
         // otherwise assume an INTERNAL_ERROR.
@@ -240,65 +284,64 @@ impl Error {
         Error::new(Kind::Canceled)
     }
 
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     pub(super) fn new_incomplete() -> Error {
         Error::new(Kind::IncompleteMessage)
     }
 
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     pub(super) fn new_too_large() -> Error {
         Error::new(Kind::Parse(Parse::TooLarge))
     }
 
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     pub(super) fn new_version_h2() -> Error {
         Error::new(Kind::Parse(Parse::VersionH2))
     }
 
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     pub(super) fn new_unexpected_message() -> Error {
         Error::new(Kind::UnexpectedMessage)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     pub(super) fn new_io(cause: std::io::Error) -> Error {
         Error::new(Kind::Io).with(cause)
     }
 
-    #[cfg(all(any(feature = "http1", feature = "http2"), feature = "tcp"))]
-    #[cfg(feature = "server")]
-    pub(super) fn new_listen<E: Into<Cause>>(cause: E) -> Error {
-        Error::new(Kind::Listen).with(cause)
-    }
-
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "server")]
-    pub(super) fn new_accept<E: Into<Cause>>(cause: E) -> Error {
-        Error::new(Kind::Accept).with(cause)
-    }
-
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    pub(super) fn new_connect<E: Into<Cause>>(cause: E) -> Error {
-        Error::new(Kind::Connect).with(cause)
-    }
-
+    #[cfg(any(
+        all(feature = "http1", any(feature = "client", feature = "server")),
+        all(feature = "http2", feature = "client")
+    ))]
     pub(super) fn new_closed() -> Error {
         Error::new(Kind::ChannelClosed)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2", feature = "stream"))]
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     pub(super) fn new_body<E: Into<Cause>>(cause: E) -> Error {
         Error::new(Kind::Body).with(cause)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     pub(super) fn new_body_write<E: Into<Cause>>(cause: E) -> Error {
         Error::new(Kind::BodyWrite).with(cause)
     }
 
+    #[cfg(any(
+        all(feature = "http1", any(feature = "client", feature = "server")),
+        feature = "ffi"
+    ))]
     pub(super) fn new_body_write_aborted() -> Error {
-        Error::new(Kind::BodyWriteAborted)
+        Error::new(Kind::User(User::BodyWriteAborted))
     }
 
     fn new_user(user: User) -> Error {
@@ -311,16 +354,9 @@ impl Error {
         Error::new_user(User::UnexpectedHeader)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    pub(super) fn new_user_unsupported_version() -> Error {
-        Error::new_user(User::UnsupportedVersion)
-    }
-
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    pub(super) fn new_user_unsupported_request_method() -> Error {
-        Error::new_user(User::UnsupportedRequestMethod)
+    #[cfg(all(feature = "http1", feature = "server"))]
+    pub(super) fn new_header_timeout() -> Error {
+        Error::new(Kind::HeaderTimeout)
     }
 
     #[cfg(feature = "http1")]
@@ -329,38 +365,32 @@ impl Error {
         Error::new_user(User::UnsupportedStatusCode)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "client")]
-    pub(super) fn new_user_absolute_uri_required() -> Error {
-        Error::new_user(User::AbsoluteUriRequired)
-    }
-
     pub(super) fn new_user_no_upgrade() -> Error {
         Error::new_user(User::NoUpgrade)
     }
 
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     pub(super) fn new_user_manual_upgrade() -> Error {
         Error::new_user(User::ManualUpgrade)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2"))]
-    #[cfg(feature = "server")]
-    pub(super) fn new_user_make_service<E: Into<Cause>>(cause: E) -> Error {
-        Error::new_user(User::MakeService).with(cause)
-    }
-
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    #[cfg(any(
+        all(any(feature = "client", feature = "server"), feature = "http1"),
+        all(feature = "server", feature = "http2")
+    ))]
     pub(super) fn new_user_service<E: Into<Cause>>(cause: E) -> Error {
         Error::new_user(User::Service).with(cause)
     }
 
-    #[cfg(any(feature = "http1", feature = "http2"))]
+    #[cfg(all(
+        any(feature = "client", feature = "server"),
+        any(feature = "http1", feature = "http2")
+    ))]
     pub(super) fn new_user_body<E: Into<Cause>>(cause: E) -> Error {
         Error::new_user(User::Body).with(cause)
     }
 
-    #[cfg(feature = "http1")]
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
     pub(super) fn new_shutdown(cause: std::io::Error) -> Error {
         Error::new(Kind::Shutdown).with(cause)
     }
@@ -370,7 +400,12 @@ impl Error {
         Error::new_user(User::AbortedByCallback)
     }
 
-    #[cfg(feature = "http2")]
+    #[cfg(all(feature = "client", any(feature = "http1", feature = "http2")))]
+    pub(super) fn new_user_dispatch_gone() -> Error {
+        Error::new(Kind::User(User::DispatchGone))
+    }
+
+    #[cfg(all(any(feature = "client", feature = "server"), feature = "http2"))]
     pub(super) fn new_h2(cause: ::h2::Error) -> Error {
         if cause.is_io() {
             Error::new_io(cause.into_io().expect("h2::Error::is_io"))
@@ -382,79 +417,94 @@ impl Error {
     fn description(&self) -> &str {
         match self.inner.kind {
             Kind::Parse(Parse::Method) => "invalid HTTP method parsed",
-            Kind::Parse(Parse::Version) => "invalid HTTP version parsed",
             #[cfg(feature = "http1")]
+            Kind::Parse(Parse::Version) => "invalid HTTP version parsed",
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::Parse(Parse::VersionH2) => "invalid HTTP version parsed (found HTTP2 preface)",
             Kind::Parse(Parse::Uri) => "invalid URI",
-            Kind::Parse(Parse::Header(Header::Token)) => "invalid HTTP header parsed",
+            #[cfg(all(feature = "http1", feature = "server"))]
+            Kind::Parse(Parse::UriTooLong) => "URI too long",
             #[cfg(feature = "http1")]
+            Kind::Parse(Parse::Header(Header::Token)) => "invalid HTTP header parsed",
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::Parse(Parse::Header(Header::ContentLengthInvalid)) => {
                 "invalid content-length parsed"
             }
-            #[cfg(feature = "http1")]
+            #[cfg(all(feature = "http1", feature = "server"))]
             Kind::Parse(Parse::Header(Header::TransferEncodingInvalid)) => {
                 "invalid transfer-encoding parsed"
             }
-            #[cfg(feature = "http1")]
+            #[cfg(all(feature = "http1", any(feature = "client", feature = "server")))]
             Kind::Parse(Parse::Header(Header::TransferEncodingUnexpected)) => {
                 "unexpected transfer-encoding parsed"
             }
+            #[cfg(any(feature = "http1", feature = "http2"))]
             Kind::Parse(Parse::TooLarge) => "message head is too large",
             Kind::Parse(Parse::Status) => "invalid HTTP status-code parsed",
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::Parse(Parse::Internal) => {
                 "internal error inside Hyper and/or its dependencies, please report"
             }
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::IncompleteMessage => "connection closed before message completed",
-            #[cfg(feature = "http1")]
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::UnexpectedMessage => "received unexpected message from connection",
+            #[cfg(any(
+                all(feature = "http1", any(feature = "client", feature = "server")),
+                all(feature = "http2", feature = "client")
+            ))]
             Kind::ChannelClosed => "channel closed",
-            Kind::Connect => "error trying to connect",
             Kind::Canceled => "operation was canceled",
-            #[cfg(all(any(feature = "http1", feature = "http2"), feature = "tcp"))]
-            #[cfg(feature = "server")]
-            Kind::Listen => "error creating server listener",
-            #[cfg(any(feature = "http1", feature = "http2"))]
-            #[cfg(feature = "server")]
-            Kind::Accept => "error accepting connection",
-            #[cfg(any(feature = "http1", feature = "http2", feature = "stream"))]
+            #[cfg(all(feature = "http1", feature = "server"))]
+            Kind::HeaderTimeout => "read header from client timeout",
+            #[cfg(all(
+                any(feature = "client", feature = "server"),
+                any(feature = "http1", feature = "http2")
+            ))]
             Kind::Body => "error reading a body from connection",
-            #[cfg(any(feature = "http1", feature = "http2"))]
+            #[cfg(all(
+                any(feature = "client", feature = "server"),
+                any(feature = "http1", feature = "http2")
+            ))]
             Kind::BodyWrite => "error writing a body to connection",
-            Kind::BodyWriteAborted => "body write aborted",
-            #[cfg(feature = "http1")]
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::Shutdown => "error shutting down connection",
-            #[cfg(feature = "http2")]
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http2"))]
             Kind::Http2 => "http2 error",
-            #[cfg(any(feature = "http1", feature = "http2"))]
+            #[cfg(all(
+                any(feature = "client", feature = "server"),
+                any(feature = "http1", feature = "http2")
+            ))]
             Kind::Io => "connection error",
 
-            #[cfg(any(feature = "http1", feature = "http2"))]
-            Kind::User(User::Body) => "error from user's HttpBody stream",
-            #[cfg(any(feature = "http1", feature = "http2"))]
-            #[cfg(feature = "server")]
-            Kind::User(User::MakeService) => "error from user's MakeService",
-            #[cfg(any(feature = "http1", feature = "http2"))]
+            #[cfg(all(
+                any(feature = "client", feature = "server"),
+                any(feature = "http1", feature = "http2")
+            ))]
+            Kind::User(User::Body) => "error from user's Body stream",
+            #[cfg(any(
+                all(feature = "http1", any(feature = "client", feature = "server")),
+                feature = "ffi"
+            ))]
+            Kind::User(User::BodyWriteAborted) => "user body write aborted",
+            #[cfg(any(
+                all(any(feature = "client", feature = "server"), feature = "http1"),
+                all(feature = "server", feature = "http2")
+            ))]
             Kind::User(User::Service) => "error from user's Service",
             #[cfg(any(feature = "http1", feature = "http2"))]
             #[cfg(feature = "server")]
             Kind::User(User::UnexpectedHeader) => "user sent unexpected header",
-            #[cfg(any(feature = "http1", feature = "http2"))]
-            #[cfg(feature = "client")]
-            Kind::User(User::UnsupportedVersion) => "request has unsupported HTTP version",
-            #[cfg(any(feature = "http1", feature = "http2"))]
-            #[cfg(feature = "client")]
-            Kind::User(User::UnsupportedRequestMethod) => "request has unsupported HTTP method",
             #[cfg(feature = "http1")]
             #[cfg(feature = "server")]
             Kind::User(User::UnsupportedStatusCode) => {
                 "response has 1xx status code, not supported by server"
             }
-            #[cfg(any(feature = "http1", feature = "http2"))]
-            #[cfg(feature = "client")]
-            Kind::User(User::AbsoluteUriRequired) => "client requires absolute-form URIs",
             Kind::User(User::NoUpgrade) => "no upgrade available",
-            #[cfg(feature = "http1")]
+            #[cfg(all(any(feature = "client", feature = "server"), feature = "http1"))]
             Kind::User(User::ManualUpgrade) => "upgrade expected but low level API in use",
+            #[cfg(all(feature = "client", any(feature = "http1", feature = "http2")))]
+            Kind::User(User::DispatchGone) => "dispatch task is gone",
             #[cfg(feature = "ffi")]
             Kind::User(User::AbortedByCallback) => "operation aborted by an application callback",
         }
@@ -474,11 +524,7 @@ impl fmt::Debug for Error {
 
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        if let Some(ref cause) = self.inner.cause {
-            write!(f, "{}: {}", self.description(), cause)
-        } else {
-            f.write_str(self.description())
-        }
+        f.write_str(self.description())
     }
 }
 
@@ -500,19 +546,23 @@ impl From<Parse> for Error {
 
 #[cfg(feature = "http1")]
 impl Parse {
+    #[cfg(any(feature = "client", feature = "server"))]
     pub(crate) fn content_length_invalid() -> Self {
         Parse::Header(Header::ContentLengthInvalid)
     }
 
+    #[cfg(feature = "server")]
     pub(crate) fn transfer_encoding_invalid() -> Self {
         Parse::Header(Header::TransferEncodingInvalid)
     }
 
+    #[cfg(any(feature = "client", feature = "server"))]
     pub(crate) fn transfer_encoding_unexpected() -> Self {
         Parse::Header(Header::TransferEncodingUnexpected)
     }
 }
 
+#[cfg(feature = "http1")]
 impl From<httparse::Error> for Parse {
     fn from(err: httparse::Error) -> Parse {
         match err {
